from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Optional, Tuple, Union

import cv2
import numpy as np
import screeninfo
from matplotlib import colors
from numpy.typing import NDArray

from ..utils._checks import _check_type
from ..utils._docs import fill_doc
from ..utils._logs import logger


@fill_doc
class BaseVisual(ABC):
    """Base visual class.

    Parameters
    ----------
    %(visual_window_name)s
    %(visual_window_size)s
    """

    @abstractmethod
    def __init__(
        self,
        window_name: str = "Visual",
        window_size: Optional[Tuple[int, int]] = None,
    ):
        _check_type(window_name, (str,), "window_name")

        self._window_name = window_name

        # size attributes
        self._window_size = BaseVisual._check_window_size(window_size)
        self._window_width = self._window_size[0]
        self._window_height = self._window_size[1]
        self._window_center = (
            self._window_width // 2,
            self._window_height // 2,
        )

        # default to black background
        self._img = np.full(
            (self._window_height, self._window_width, 3),
            fill_value=(0, 0, 0),
            dtype=np.uint8,
        )
        self._background = (0, 0, 0)

    def show(self, wait: int = 1) -> None:
        """Show the visual with cv2.imshow() and cv2.waitKey().

        Parameters
        ----------
        wait : int
            Wait timer passed to cv2.waitKey() [ms].
        """
        _check_type(wait, ("int",), "wait")
        cv2.imshow(self._window_name, self._img)
        cv2.waitKey(wait)

    def close(self) -> None:
        """Close the visual."""
        cv2.destroyWindow(self._window_name)

    @fill_doc
    def draw_background(self, color: Union[str, Tuple[int, int, int]]) -> None:
        """Draw a uniform single color background.

        Replace all the pixels with this color, thus this method erases any
        prior work.

        Parameters
        ----------
        color : str | tuple
            Color used to draw the background. %(visual_color)s
        """
        color = BaseVisual._check_color(color)
        self._img = np.full(
            (self._window_height, self._window_width, 3),
            fill_value=color,
            dtype=np.uint8,
        )
        self._background = color

    def __del__(self):
        """Close when deleting the object."""
        self.close()

    # --------------------------------------------------------------------
    @staticmethod
    def _check_window_size(
        window_size: Optional[Tuple[int, int]]
    ) -> Tuple[int, int]:
        """Check if the window size is valid.

        If None, set it as the minimum (width, height) supported by any
        connected monitor.
        """
        _check_type(window_size, (None, tuple), "window_size")

        if window_size is None:
            try:
                width = min(
                    monitor.width for monitor in screeninfo.get_monitors()
                )
                height = min(
                    monitor.height for monitor in screeninfo.get_monitors()
                )
            except Exception as error:
                logger.error("No monitor found.")
                raise error
            window_size = (width, height)

        for size in window_size:
            _check_type(size, ("int",))
        assert len(window_size) == 2
        assert all(0 < size for size in window_size)
        return window_size

    @staticmethod
    def _check_color(
        color: Union[str, Tuple[int, int, int]]
    ) -> Tuple[int, int, int]:
        """Check if a color is valid and converts it to BGR."""
        _check_type(color, (str, tuple), "color")
        if isinstance(color, str):
            r, g, b, _ = colors.to_rgba(color)
            color = tuple([int(c * 255) for c in (b, g, r)])
        assert len(color) == 3
        assert all(0 <= c <= 255 for c in color)
        return color

    # --------------------------------------------------------------------
    @property
    def window_name(self) -> str:
        """Window's name."""
        return self._window_name

    @property
    def window_size(self) -> Tuple[int, int]:
        """Window's size (width x height)."""
        return self._window_size

    @property
    def window_center(self) -> Tuple[int, int]:
        """Window's center position."""
        return self._window_center

    @property
    def img(self) -> NDArray[int]:
        """Image array."""
        return self._img

    @property
    def background(self) -> Tuple[int, int, int]:
        """Background color in BGR color space."""
        return self._background

    @background.setter
    def background(self, background):
        self.draw_background(background)


class BaseFeedbackVisual(BaseVisual):
    """Base visual feedback class.

    Parameters
    ----------
    %(visual_window_name)s
    %(visual_window_size)s
    """

    @abstractmethod
    def __init__(
        self,
        window_name: str = "Visual",
        window_size: Optional[Tuple[int, int]] = None,
    ):
        super().__init__(window_name, window_size)
        self._backup_img = None

    def _reset(self):
        """Reset the visual with the backup, thus removing the feedback."""
        if self._backup_img is not None:
            self._img = deepcopy(self._backup_img)

    # --------------------------------------------------------------------
    @staticmethod
    def _check_axis(axis: Union[str, int]) -> int:
        """Check that the axis is valid and converts it to integer (0, 1).

        The axis is a binary integer:
        - 0: Vertical
        - 1: Horizontal
        """
        _check_type(axis, ("int", str), "axis")
        if isinstance(axis, str):
            axis = axis.lower().strip()
            assert axis in ["horizontal", "h", "vertical", "v"]
            axis = 0 if axis.startswith("v") else 1
        assert axis in (0, 1)
        return axis



from itertools import product

import numpy as np
import pytest
from scipy.signal import find_peaks

from .. import SoundAM
from .test_base import _test_base, _test_no_volume


def _check_frequency(signal, sample_rate, carrier, modulation, method):
    """Check frequency of the ASSR with conventional modulation."""
    frequencies = np.fft.rfftfreq(signal.shape[0], 1 / sample_rate)
    fftval = np.abs(np.fft.rfft(signal, axis=0))
    height = 0.1 * np.max(fftval)
    peaks1 = find_peaks(fftval[:, 0], height=height)[0]
    peaks1 = peaks1.astype(int)
    peaks2 = find_peaks(fftval[:, 1], height=height)[0]
    peaks2 = peaks2.astype(int)
    assert (peaks1 == peaks2).all()
    if method == "conventional":
        peaks = peaks1
        assert peaks.size == 3
        idx = np.argmax(fftval, axis=0)
        assert np.allclose(frequencies[idx], np.array([carrier, carrier]))
        # remove carrier from peaks
        peaks = [elt for elt in peaks1 if elt not in idx]
    elif method == "dsbsc":
        peaks = peaks1
        assert peaks.size == 2
        assert carrier not in frequencies[peaks]
    assert np.allclose(
        frequencies[peaks],
        np.array([carrier - modulation, carrier + modulation]),
    )


@pytest.mark.parametrize(
    "volume, sample_rate, duration, method",
    product((10, 100), (44100, 48000), (1, 5), ("conventional", "dsbsc")),
)
def test_assr(volume, sample_rate, duration, method):
    """Test an AM sound."""
    sound = SoundAM(volume, sample_rate, duration, method=method)
    assert sound.volume == volume
    assert sound.sample_rate == sample_rate
    assert sound.duration == duration
    assert sound.times.size == sound.sample_rate * sound.duration
    assert sound.signal.shape == (sound.times.size, 2)
    assert np.isclose(np.max(np.abs(sound.signal)), volume / 100)
    assert sound.frequency_carrier == 1000
    assert sound.frequency_modulation == 40
    _check_frequency(sound.signal, sound.sample_rate, 1000, 40, method)


@pytest.mark.parametrize(
    "carrier, modulation, method",
    product((300, 1000, 5000), (20, 50), ("conventional", "dsbsc")),
)
def test_assr_frequencies(carrier, modulation, method):
    """Test AM with different frequencies."""
    sound = SoundAM(
        volume=10,
        frequency_carrier=carrier,
        frequency_modulation=modulation,
        method=method,
    )
    assert sound.frequency_carrier == carrier
    assert sound.frequency_modulation == modulation
    _check_frequency(
        sound.signal,
        sound.sample_rate,
        carrier,
        modulation,
        method,
    )


@pytest.mark.parametrize(
    "carrier, modulation, method",
    product((300, 1000, 5000), (20, 50), ("conventional", "dsbsc")),
)
def test_assr_frequencies_setter(carrier, modulation, method):
    """Test AM frequency setter."""
    sound = SoundAM(
        volume=10,
        frequency_carrier=200,
        frequency_modulation=10,
        method=method,
    )
    assert sound.frequency_carrier == 200
    assert sound.frequency_modulation == 10
    sound.frequency_carrier = carrier
    assert sound.frequency_carrier == carrier
    assert sound.frequency_modulation == 10
    _check_frequency(
        sound.signal,
        sound.sample_rate,
        carrier,
        sound.frequency_modulation,
        method,
    )
    sound.frequency_modulation = modulation
    assert sound.frequency_carrier == carrier
    assert sound.frequency_modulation == modulation
    _check_frequency(
        sound.signal,
        sound.sample_rate,
        carrier,
        modulation,
        method,
    )


def test_assr_method_setter():
    """Test AM method setter."""
    method = "conventional"
    sound = SoundAM(10, method=method)
    assert sound.method == method
    _check_frequency(sound.signal, sound.sample_rate, 1000, 40, method)
    method = "dsbsc"
    sound.method = method
    assert sound.method == method
    _check_frequency(sound.signal, sound.sample_rate, 1000, 40, method)
    method = "conventional"
    sound.method = method
    assert sound.method == method
    _check_frequency(sound.signal, sound.sample_rate, 1000, 40, method)


def test_base():
    """Test base functionalities with an AM sound."""
    _test_base(SoundAM)


def test_no_volume():
    """Test signal if volume is set to 0."""
    _test_no_volume(SoundAM)



from itertools import product

import numpy as np
import pytest

from .. import MovingBar
from .test_base import _test_base


@pytest.mark.parametrize(
    "length, width, color, position, axis",
    product(
        (100, 200),
        (10, 20),
        ("white", (101, 101, 101)),
        (-1, -0.5, 0, 0.5, 1),
        (0, 1, "v", "h"),
    ),
)
def test_moving_bar(length, width, color, position, axis):
    """Test a moving bar visual."""
    visual = MovingBar(window_name="test", window_size=(500, 500))
    assert visual.window_name == "test"
    assert visual.window_size == (500, 500)
    assert visual._window_center == (250, 250)
    assert np.count_nonzero(visual.img) == 0
    visual.putBar(length, width, color, position, axis)


def test_moving_bar_setters():
    """Test property setters."""
    visual = MovingBar(window_name="test", window_size=(500, 500))
    visual.putBar(100, 20, "white", 0, 0)

    assert visual.position == 0
    visual.position = 0.5
    assert visual.position == 0.5

    assert visual.color == (255, 255, 255)
    visual.color = (210, 210, 210)
    assert visual.color == (210, 210, 210)

    assert visual.length == 100
    visual.length = 200
    assert visual.length == 200

    assert visual.width == 20
    visual.width = 30
    assert visual.width == 30

    assert visual.axis == 0
    visual.axis = 1
    assert visual.axis == 1


def test_base():
    """Test base functionalities with a moving bar visual."""
    _test_base(MovingBar)


def test_invalid_arguments():
    """Test invalid arguments."""
    visual = MovingBar(window_name="test", window_size=(500, 500))

    with pytest.raises(TypeError, match="must be an instance of"):
        visual.putBar("long", 20, "white", 0, 0)
    with pytest.raises(TypeError, match="must be an instance of"):
        visual.putBar(100, "long", "white", 0, 0)
    with pytest.raises(TypeError, match="must be an instance of"):
        visual.putBar(100, 20, "white", "center", 0)

    with pytest.raises(AssertionError):
        visual.putBar(-100, 20, "white", 0, 0)
    with pytest.raises(AssertionError):
        visual.putBar(100, -20, "white", 0, 0)
    with pytest.raises(AssertionError):
        visual.putBar(100, 20, "white", 10, 0)



"""Base class for sound delivery."""

from abc import ABC, abstractmethod
from os import makedirs
from pathlib import Path
from typing import Tuple, Union

import numpy as np
import sounddevice as sd
from numpy.typing import NDArray
from scipy.io import wavfile

from .. import logger
from ..utils._checks import _check_type
from ..utils._docs import fill_doc


@fill_doc
class BaseSound(ABC):
    """Base audio stimulus class.

    Parameters
    ----------
    %(audio_volume)s
    %(audio_sample_rate)s
    %(audio_duration)s
    """

    @abstractmethod
    def __init__(
        self,
        volume: Union[float, Tuple[float, float]],
        sample_rate: int = 44100,
        duration: float = 1,
    ):
        self._volume = BaseSound._check_volume(volume)
        self._sample_rate = BaseSound._check_sample_rate(sample_rate)
        self._duration = BaseSound._check_duration(duration)
        self._set_times()
        self._set_signal()

    def _set_times(self) -> None:
        """Update the time array annd the ._signal variable."""
        logger.debug(
            "Setting the 'times' array with the duration %.2f "
            "[seconds] and the sampling rate %.1f [Hz].",
            self._duration,
            self._sample_rate,
        )
        self._times = np.linspace(
            0,
            self._duration,
            int(self._duration * self._sample_rate),
            endpoint=True,
        )

    @abstractmethod
    def _set_signal(self) -> None:
        """Set the signal in the numpy array ._signal played by sounddevice."""
        # [:, 0] for left and [:, 1] for right
        self._signal = np.zeros(shape=(self._times.size, 2))

    # --------------------------------------------------------------------
    def play(self, blocking: bool = False) -> None:
        """Play the sound.

        This function creates and terminates an audio stream.
        """
        _check_type(blocking, (bool,), "blocking")
        logger.debug("Play requested with blocking set to %s.", blocking)
        sd.play(
            self._signal,
            samplerate=self._sample_rate,
            mapping=[1, 2],
            blocking=blocking,
        )

    def stop(self) -> None:
        """Stop the sounds played on the active audio stream."""
        logger.debug("Stop requested.")
        sd.stop()

    def save(self, fname: Union[str, Path], overwrite: bool = False) -> None:
        """Save a sound signal into a .wav file with scipy.io.wavfile.write().

        Parameters
        ----------
        fname : str | Path
            Path to the file where the sound signal is saved. The extension
            should be '.wav'.
        overwrite : bool
            If True, file with the same name are overwritten.
        """
        fname = BaseSound._check_file(fname, must_exists=False)
        if overwrite is False and fname.exists():
            raise RuntimeError(
                "The file %s already exist. Set argument "
                "'overwrite' to True if you want to overwrite "
                "the existing file.",
                fname,
            )
        if not fname.parent.exists():
            makedirs(fname.parent)
        logger.debug(
            "Writing sound to file %s with sampling frequency %.1f [Hz].",
            fname,
            self._sample_rate,
        )
        wavfile.write(fname, self._sample_rate, self._signal)

    # --------------------------------------------------------------------
    @staticmethod
    def _check_volume(
        volume: Union[float, Tuple[float, float]]
    ) -> NDArray[float]:
        """Check that the volume provided by the user is valid."""
        _check_type(volume, ("numeric", tuple), "volume")
        if not isinstance(volume, tuple):
            volume = (volume, volume)
        assert len(volume) in (1, 2)
        for vol in volume:
            _check_type(vol, ("numeric",))
        assert all(0 <= v <= 100 for v in volume)
        return np.array(volume)

    @staticmethod
    def _check_sample_rate(sample_rate: int) -> int:
        """Check if the sample rate is a positive integer."""
        _check_type(sample_rate, ("int",), item_name="sample_rate")
        assert 0 < sample_rate
        return sample_rate

    @staticmethod
    def _check_duration(duration: float) -> float:
        """Check if the duration is positive."""
        _check_type(duration, ("numeric",), item_name="duration")
        assert 0 < duration
        return duration

    @staticmethod
    def _check_file(fname: Union[str, Path], must_exists: bool) -> Path:
        """Check if the fname is valid."""
        SUPPORTED = ".wav"

        _check_type(fname, ("path-like",))
        fname = Path(fname)
        assert fname.suffix in SUPPORTED
        if must_exists:
            assert fname.exists()
        return fname

    # --------------------------------------------------------------------
    @property
    def volume(self) -> Union[float, Tuple[float, float]]:
        """Sound's volume(s) [AU]."""
        logger.debug("'self._volume' is set to %s [AU].", self._volume)
        volume = tuple(self._volume)
        if volume[0] == volume[1]:
            volume = volume[0]
        return volume

    @volume.setter
    def volume(self, volume: Union[float, Tuple[float, float]]):
        logger.debug("Setting 'volume' to %s [AU].", volume)
        self._volume = BaseSound._check_volume(volume)
        self._set_signal()

    @property
    def sample_rate(self) -> int:
        """Sound's sampling rate [Hz]."""
        logger.debug(
            "'self._sample_rate' is set to %.1f [Hz].", self._sample_rate
        )
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, sample_rate: int):
        logger.debug("Setting 'sample_rate' to %.1f [Hz].", sample_rate)
        self._sample_rate = BaseSound._check_sample_rate(sample_rate)
        self._set_times()
        self._set_signal()

    @property
    def duration(self) -> float:
        """Sound's duration [seconds]."""
        logger.debug(
            "'self._duration' is set to %.2f [seconds].", self._duration
        )
        return self._duration

    @duration.setter
    def duration(self, duration: float):
        logger.debug("Setting 'duration' to %.2f [seconds].", duration)
        self._duration = BaseSound._check_duration(duration)
        self._set_times()
        self._set_signal()

    @property
    def signal(self) -> NDArray[float]:
        """Sound's signal."""
        return self._signal

    @property
    def times(self) -> NDArray[float]:
        """Times array."""
        return self._times



import numpy as np
import pytest

from stimuli.visuals import BaseFeedbackVisual, BaseVisual


def test_check_window_size():
    """Check the window size static checker."""
    winsize = BaseVisual._check_window_size((100, 20))
    assert winsize == (100, 20)

    with pytest.raises(TypeError, match="must be an instance of"):
        BaseVisual._check_window_size([100, 20])
    with pytest.raises(AssertionError):
        BaseVisual._check_window_size((100, 20, 101))
    with pytest.raises(AssertionError):
        BaseVisual._check_window_size((100, -20))


def test_check_color():
    """Check the color static checker."""
    color = BaseVisual._check_color("teal")
    assert isinstance(color, tuple)
    assert len(color) == 3
    assert all(isinstance(c, int) for c in color)
    assert all(0 <= c <= 255 for c in color)

    color = BaseVisual._check_color((100, 20, 101))
    assert color == (100, 20, 101)

    with pytest.raises(TypeError, match="must be an instance of"):
        BaseVisual._check_color([100, 20, 101])
    with pytest.raises(AssertionError):
        BaseVisual._check_color((100, 101))
    with pytest.raises(AssertionError):
        BaseVisual._check_color((101, -101, 101))


def test_check_axis():
    """Check the axis static checker."""
    for axis in (0, "vertical", "v"):
        ax = BaseFeedbackVisual._check_axis(axis)
        assert ax == 0
    for axis in (1, "horizontal", "h"):
        ax = BaseFeedbackVisual._check_axis(axis)
        assert ax == 1

    with pytest.raises(TypeError, match="must be an instance of"):
        BaseFeedbackVisual._check_axis([0])
    with pytest.raises(TypeError, match="must be an instance of"):
        BaseFeedbackVisual._check_axis(True)
    with pytest.raises(AssertionError):
        BaseFeedbackVisual._check_axis("101")


def _test_base(Visual):
    """Test base functionalities with a Visual class."""
    visual = Visual(window_size=(100, 50))
    assert np.all(visual.img == 0)
    visual.draw_background(color=(101, 101, 101))
    assert np.all(visual.img == 101)
    assert visual.window_name == "Visual"
    assert visual.window_size == (100, 50)
    assert visual._window_width == 100
    assert visual._window_height == 50
    assert visual.window_center == (50, 25)
    visual.background = (50, 50, 50)
    assert np.all(visual.img == 50)



import logging
import sys
from typing import Callable, Optional, Union

from ._checks import _check_verbose
from ._docs import fill_doc

logger = logging.getLogger(__package__.split(".utils", maxsplit=1)[0])
logger.propagate = False  # don't propagate (in case of multiple imports)


@fill_doc
def init_logger(verbose: Optional[Union[bool, str, int]] = None) -> None:
    """Initialize a logger.

    Assign sys.stdout as a handler of the logger.

    Parameters
    ----------
    %(verbose)s
    """
    set_log_level(verbose)
    add_stream_handler(sys.stdout, verbose)


@fill_doc
def add_stream_handler(
    stream, verbose: Optional[Union[bool, str, int]] = None
) -> None:
    """Add a stream handler to the logger.

    Parameters
    ----------
    stream : The output stream, e.g. sys.stdout
    %(verbose)s
    """
    verbose = _check_verbose(verbose)
    handler = logging.StreamHandler(stream)
    handler.setFormatter(LoggerFormatter())
    logger.addHandler(handler)
    set_handler_log_level(-1, verbose)


@fill_doc
def add_file_handler(
    fname, mode: str = "a", verbose: Optional[Union[bool, str, int]] = None
) -> None:
    """Add a file handler to the logger.

    Parameters
    ----------
    fname : str | Path
    mode : str
        Mode in which the file is opened.
    %(verbose)s
    """
    verbose = _check_verbose(verbose)
    handler = logging.FileHandler(fname, mode)
    handler.setFormatter(LoggerFormatter())
    logger.addHandler(handler)
    set_handler_log_level(-1, verbose)


@fill_doc
def set_handler_log_level(
    handler_id: int, verbose: Union[bool, str, int, None]
) -> None:
    """Set the log level for a specific handler.

    First handler (ID 0) is always stdout, followed by user-defined handlers.

    Parameters
    ----------
    handler_id : int
        ID of the handler among 'logger.handlers'.
    %(verbose)s
    """
    verbose = _check_verbose(verbose)
    logger.handlers[handler_id].setLevel = verbose


@fill_doc
def set_log_level(verbose: Union[bool, str, int, None]) -> None:
    """Set the log level for the logger.

    Parameters
    ----------
    %(verbose)s
    """
    verbose = _check_verbose(verbose)
    logger.setLevel(verbose)


class LoggerFormatter(logging.Formatter):
    """Format string Syntax."""

    # Format string syntax for the different Log levels
    _formatters = dict()
    _formatters[logging.DEBUG] = logging.Formatter(
        fmt="[%(module)s:%(funcName)s:%(lineno)d] %(levelname)s: %(message)s "
        "(%(asctime)s)"
    )
    _formatters[logging.INFO] = logging.Formatter(
        fmt="[%(module)s.%(funcName)s] %(levelname)s: %(message)s"
    )
    _formatters[logging.WARNING] = logging.Formatter(
        fmt="[%(module)s.%(funcName)s] %(levelname)s: %(message)s"
    )
    _formatters[logging.ERROR] = logging.Formatter(
        fmt="[%(module)s:%(funcName)s:%(lineno)d] %(levelname)s: %(message)s"
    )

    def __init__(self):
        super().__init__(fmt="%(levelname): %(message)s")

    def format(self, record: logging.LogRecord):
        """
        Format the received log record.

        Parameters
        ----------
        record : logging.LogRecord
        """
        if record.levelno <= logging.DEBUG:
            return self._formatters[logging.DEBUG].format(record)
        elif record.levelno <= logging.INFO:
            return self._formatters[logging.INFO].format(record)
        elif record.levelno <= logging.WARNING:
            return self._formatters[logging.WARNING].format(record)
        else:
            return self._formatters[logging.ERROR].format(record)


def verbose(f: Callable) -> Callable:
    """Set the verbose for the function call from the kwargs.

    Parameters
    ----------
    f : callable
        The function with a verbose argument.

    Returns
    -------
    f : callable
        The function.
    """

    def wrapper(*args, **kwargs):
        if "verbose" in kwargs:
            set_log_level(kwargs["verbose"])
        return f(*args, **kwargs)

    return wrapper


init_logger()



from itertools import product

import numpy as np
import pytest

from .. import Tone
from .test_base import _test_base, _test_no_volume


def _check_frequency(signal, sample_rate, target):
    """Check frequency of the tone."""
    frequencies = np.fft.rfftfreq(signal.shape[0], 1 / sample_rate)
    fftval = np.abs(np.fft.rfft(signal, axis=0))
    idx = np.argmax(fftval, axis=0)
    assert np.allclose(frequencies[idx], np.array([target, target]))


@pytest.mark.parametrize(
    "volume, sample_rate, duration", product((10, 100), (44100, 48000), (1, 5))
)
def test_tone(volume, sample_rate, duration):
    """Test a tone sound."""
    sound = Tone(volume, sample_rate, duration)
    assert sound.volume == volume
    assert sound.sample_rate == sample_rate
    assert sound.duration == duration
    assert sound.times.size == sound.sample_rate * sound.duration
    assert sound.signal.shape == (sound.times.size, 2)
    assert np.isclose(np.max(np.abs(sound.signal)), volume / 100)
    assert sound.frequency == 440
    _check_frequency(sound.signal, sound.sample_rate, 440)


@pytest.mark.parametrize("frequency", (101, 440, 1000))
def test_tone_frequency(frequency):
    """Test tone with different frequencies."""
    sound = Tone(volume=10, frequency=frequency)
    assert sound.frequency == frequency
    _check_frequency(sound.signal, sound.sample_rate, frequency)


@pytest.mark.parametrize("frequency", (101, 440, 1000))
def test_tone_frequency_setter(frequency):
    """Test tone frequency setter."""
    sound = Tone(volume=10, frequency=10)
    assert sound.frequency == 10
    sound.frequency = frequency
    assert sound.frequency == frequency
    _check_frequency(sound.signal, sound.sample_rate, frequency)


def test_base():
    """Test base functionalities with a tone."""
    _test_base(Tone)


def test_no_volume():
    """Test signal if volume is set to 0."""
    _test_no_volume(Tone)



"""Test _imports.py"""

import pytest

from .._imports import import_optional_dependency


def test_import_optional_dependency():
    """Test the import of optional dependencies."""
    # Test import of present package
    numpy = import_optional_dependency("numpy")
    assert isinstance(numpy.__version__, str)

    # Test import of absent package
    with pytest.raises(ImportError, match="Missing optional dependency"):
        import_optional_dependency("non_existing_pkg", raise_error=True)

    # Test import of absent package without raise
    pkg = import_optional_dependency("non_existing_pkg", raise_error=False)
    assert pkg is None

    # Test extra
    with pytest.raises(ImportError, match="blabla"):
        import_optional_dependency("non_existing_pkg", extra="blabla")



from itertools import product

import numpy as np
import pytest

from .. import Noise
from ..noise import _PSDS
from .test_base import _test_base, _test_no_volume


@pytest.mark.parametrize(
    "volume, sample_rate, duration, color",
    product((10, 100), (44100, 48000), (1, 5), _PSDS.keys()),
)
def test_noise(volume, sample_rate, duration, color):
    """Test a noise sound."""
    sound = Noise(volume, sample_rate, duration, color)
    assert sound.volume == volume
    assert sound.sample_rate == sample_rate
    assert sound.duration == duration
    assert sound.color == color
    assert sound.times.size == sound.sample_rate * sound.duration
    assert sound.signal.shape == (sound.times.size, 2)
    assert np.isclose(np.max(np.abs(sound.signal)), volume / 100)


def test_color_setter():
    """Test color setter."""
    sound = Noise(10, color="white")
    assert sound.color == "white"
    sound.color = "pink"
    assert sound.color == "pink"


def test_base():
    """Test base functionalities with a noise."""
    _test_base(Noise)


def test_no_volume():
    """Test signal if volume is set to 0."""
    _test_no_volume(Noise)



from copy import deepcopy
from typing import Optional, Tuple, Union

import cv2

from ..utils._checks import _check_type
from ..utils._docs import fill_doc
from .base import BaseFeedbackVisual


@fill_doc
class MovingBar(BaseFeedbackVisual):
    """Class to display a centered moving bar along an axis.

    Parameters
    ----------
    %(visual_window_name)s
    %(visual_window_size)s
    """

    def __init__(
        self,
        window_name: str = "Visual",
        window_size: Optional[Tuple[int, int]] = None,
    ):
        super().__init__(window_name, window_size)

    @fill_doc
    def putBar(
        self,
        length: int,
        width: int,
        color: Union[str, Tuple[int, int, int]],
        position: float = 0,
        axis: Union[int, str] = 0,
    ):
        """Draw the bar on top of the current visual.

        Parameters
        ----------
        %(visual_length)s
        %(visual_width)s
        color : str | tuple
            Color used to draw the bar. %(visual_color)s
        position : float
            Relative position of the bar along the given axis.
            Along the vertical axis:
                - -1: corresponds to the top of the window.
                - 1: corresponds to the bottom of the window.
            Along the horizontal axis:
                - -1: corresponds to the left of the window.
                - 1: corresponds to the right of the window.
            0 corresponds to the center of the window.
        axis : int | str
            Axis along which the bar is moving:
                - 0 | 'vertical' | 'v'
                    horizontal bar along vertical axis.
                - 1 | 'horizontal' | 'h'
                    vertical bar along horizontal axis.
        """
        if self._backup_img is None:
            self._backup_img = deepcopy(self._img)
        else:
            self._reset()

        self._position = MovingBar._check_position(position)
        self._axis = BaseFeedbackVisual._check_axis(axis)

        self._length = MovingBar._check_length(
            length, self._axis, self._window_size
        )
        self._width = MovingBar._check_width(
            width, self._length, self._axis, self._window_size
        )
        self._color = BaseFeedbackVisual._check_color(color)

        self._putBar()

    def _putBar(self):
        """Draw the bar rectangle.

        - Axis = 0 - Horizontal bar along vertical axis.
        P1 ---------------
        |                |
        --------------- P2

        - Axis = 1 - Vertical bar along horizontal axis
        P1 ---
        |    |
        |    |
        |    |
        |    |
        |    |
        --- P2
        """
        position = MovingBar._convert_position_to_pixel(
            self._position, self._axis, self._window_size, self.window_center
        )

        if self._axis == 0:
            xP1 = self.window_center[0] - self._length // 2
            yP1 = position - self._width // 2
            xP2 = xP1 + self._length
            yP2 = yP1 + self._width
        elif self._axis == 1:
            xP1 = position - self._width // 2
            yP1 = self.window_center[1] - self._length // 2
            xP2 = xP1 + self._width
            yP2 = yP1 + self._length

        cv2.rectangle(self._img, (xP1, yP1), (xP2, yP2), self._color, -1)

    # --------------------------------------------------------------------
    @staticmethod
    def _check_length(
        length: int, axis: int, window_size: Tuple[int, int]
    ) -> int:
        """Check that the length is valid."""
        _check_type(length, ("int",), "length")
        assert 0 < length
        assert length <= window_size[axis]
        return length

    @staticmethod
    def _check_width(
        width: int, length: int, axis: int, window_size: Tuple[int, int]
    ) -> int:
        """Check that the width is valid."""
        _check_type(width, ("int",), "width")
        assert 0 < width
        assert width <= length
        assert width <= window_size[(axis + 1) % 2]
        return width

    @staticmethod
    def _check_position(position: float) -> float:
        """Check that the position given is between -1 and 1."""
        _check_type(position, ("numeric",), "position")
        assert -1 <= position <= 1
        return position

    @staticmethod
    def _convert_position_to_pixel(
        position: float,
        axis: int,
        window_size: Tuple[int, int],
        window_center: Tuple[int, int],
    ) -> int:
        """Convert the position [-1, 1] to an absolute position in pixel."""
        idx = (axis + 1) % 2
        if position == 0:
            return window_center[idx]
        elif -1 <= position < 0:
            return int(window_center[idx] * (1 - abs(position)))
        elif 0 < position <= 1:
            return int(
                window_center[idx]
                + (window_size[idx] - window_center[idx]) * position
            )

    # --------------------------------------------------------------------
    @property
    def length(self) -> int:
        """Length of the bar in pixel."""
        return self._length

    @length.setter
    def length(self, length):
        self._length = MovingBar._check_length(
            length, self._axis, self._window_size
        )
        self._reset()
        self._putBar()

    @property
    def width(self) -> int:
        """Width of the bar in pixel."""
        return self._width

    @width.setter
    def width(self, width):
        self._width = MovingBar._check_width(
            width, self._length, self._axis, self._window_size
        )
        self._reset()
        self._putBar()

    @property
    def color(self) -> Tuple[int, int, int]:
        """Color of the bar in BGR color space."""
        return self._color

    @color.setter
    def color(self, color):
        self._color = BaseFeedbackVisual._check_color(color)
        self._reset()
        self._putBar()

    @property
    def position(self) -> float:
        """Position between -1 and 1 of the bar on the given axis."""
        return self._position

    @position.setter
    def position(self, position):
        self._position = MovingBar._check_position(position)
        self._reset()
        self._putBar()

    @property
    def axis(self):
        """Axis on which the bar is moving.

        This property is a binary integer:
            - 0: Horizontal bar along vertical axis.
            - 1: Vertical bar along horizontal axis.
        """
        return self._axis

    @axis.setter
    def axis(self, axis):
        self._axis = BaseFeedbackVisual._check_axis(axis)
        self._reset()
        self._putBar()



"""Amplitude modulated sound."""

from typing import Tuple, Union

import numpy as np

from .. import logger
from ..utils._checks import _check_type, _check_value
from ..utils._docs import copy_doc, fill_doc
from .base import BaseSound


@fill_doc
class SoundAM(BaseSound):
    """Amplitude modulated sound.

    Composed of a carrier frequency fc which is amplitude modulated at fm.
    By default, an Auditory Steady State Response stimuli composed of a 1000 Hz
    carrier frequency modulated at 40 Hz through conventional modulation is
    created.

    Parameters
    ----------
    %(audio_volume)s
    %(audio_sample_rate)s
    %(audio_duration)s
    frequency_carrier : int
        Carrier frequency in Hz.
    frequency_modulation : int
        Modulatiom frequency in Hz.
    method : str
        Either 'conventional' or 'dsbsc'.
        - 'conventional':
            Also called classical AM, the eq. used is:
                signal = (1 - M(t)) * cos(2*pi*fc*t)
                M(t) = cos(2*pi*fm*t)
        - 'dsbsc':
            Also called double side band suppressed carrier, the eq. used is:
                signal = M(t)*cos(2*pi*fc*t)
                M(t) = sin(2*pi*fm*t)
    """

    def __init__(
        self,
        volume: Union[float, Tuple[float, float]],
        sample_rate: int = 44100,
        duration: float = 1,
        frequency_carrier: float = 1000,
        frequency_modulation: float = 40,
        method: str = "conventional",
    ):
        self._frequency_carrier = SoundAM._check_frequency_carrier(
            frequency_carrier
        )
        self._frequency_modulation = SoundAM._check_frequency_modulation(
            frequency_modulation
        )
        _check_type(method, (str,), "method")
        _check_value(method, ("conventional", "dsbsc"), "AM method")
        self._method = method
        self.name = f"AM {self._method}"
        super().__init__(volume, sample_rate, duration)

    @copy_doc(BaseSound._set_signal)
    def _set_signal(self) -> None:
        if self._method == "conventional":
            amplitude = 1 - np.cos(
                2 * np.pi * self._frequency_modulation * self._times
            )
            arr = amplitude * np.cos(
                2 * np.pi * self._frequency_carrier * self._times
            )

        elif self._method == "dsbsc":
            amplitude = np.sin(
                2 * np.pi * self._frequency_modulation * self._times
            )
            arr = amplitude * np.sin(
                2 * np.pi * self._frequency_carrier * self._times
            )
        arr /= np.max(np.abs(arr))  # normalize
        self._signal = np.vstack((arr, arr)).T * self._volume / 100

    # --------------------------------------------------------------------
    @staticmethod
    def _check_frequency_carrier(frequency_carrier: float) -> float:
        """Check if the carrier frequency is positive."""
        _check_type(
            frequency_carrier, ("numeric",), item_name="frequency_carrier"
        )
        assert 0 < frequency_carrier
        return frequency_carrier

    @staticmethod
    def _check_frequency_modulation(frequency_modulation: float) -> float:
        """Check if the modulation frequency is positive."""
        _check_type(
            frequency_modulation,
            ("numeric",),
            item_name="frequency_modulation",
        )
        assert 0 < frequency_modulation
        return frequency_modulation

    # --------------------------------------------------------------------
    @property
    def frequency_carrier(self) -> float:
        """Sound's carrier frequency [Hz]."""
        logger.debug(
            "'self._frequency_carrier' is set to %.2f [Hz].",
            self._frequency_carrier,
        )
        return self._frequency_carrier

    @frequency_carrier.setter
    def frequency_carrier(self, frequency_carrier: float):
        logger.debug(
            "Setting 'frequency_carrier' to %.2f [Hz].", frequency_carrier
        )
        self._frequency_carrier = SoundAM._check_frequency_carrier(
            frequency_carrier
        )
        self._set_signal()

    @property
    def frequency_modulation(self) -> float:
        """Sound's modulation frequency [Hz]."""
        logger.debug(
            "'self._frequency_modulation' is set to %.2f [Hz].",
            self._frequency_modulation,
        )
        return self._frequency_modulation

    @frequency_modulation.setter
    def frequency_modulation(self, frequency_modulation: float):
        logger.debug(
            "Setting 'frequency_modulation' to %.2f [Hz].",
            frequency_modulation,
        )
        self._frequency_modulation = SoundAM._check_frequency_modulation(
            frequency_modulation
        )
        self._set_signal()

    @property
    def method(self) -> str:
        """Sound's modulation method."""
        logger.debug("'self._method' is set to %s.", self._method)
        return self._method

    @method.setter
    def method(self, method: str):
        logger.debug("Setting 'method' to %s.", method)
        _check_type(method, (str,), "method")
        _check_value(method, ("conventional", "dsbsc"), "AM method")
        self._method = method
        self._set_signal()



"""Colored noise sound."""

from typing import Tuple, Union

import numpy as np

from .. import logger
from ..utils._checks import _check_type, _check_value
from ..utils._docs import copy_doc, fill_doc
from .base import BaseSound

_PSDS = {
    "white": lambda f: 1,
    "blue": lambda f: np.sqrt(f),
    "violet": lambda f: f,
    "pink": lambda f: 1 / np.where(f == 0, np.inf, np.sqrt(f)),
    "brown": lambda f: 1 / np.where(f == 0, np.inf, f),
}


@fill_doc
class Noise(BaseSound):  # noqa: E501
    """Colored noise stimulus.

    This class is based on the StackOverflow answer from Bob:
    https://stackoverflow.com/questions/67085963/generate-colors-of-noise-in-python/67127726#67127726

    Parameters
    ----------
    %(audio_volume)s
    %(audio_sample_rate)s
    %(audio_duration)s
    color : str
        The noise color. Available colors are: "white", "pink",
        "blue", "violet" and "brown".
    """

    def __init__(
        self,
        volume: Union[float, Tuple[float, float]],
        sample_rate: int = 44100,
        duration: float = 1,
        color: str = "white",
    ):
        _check_type(color, (str,), "color")
        _check_value(color, _PSDS)
        self._color = color
        self._rng = np.random.default_rng()
        self.name = f"{color} noise"
        super().__init__(volume, sample_rate, duration)

    @copy_doc(BaseSound._set_signal)
    def _set_signal(self) -> None:
        noise_arr = Noise._noise_psd(self._rng, self._times.size, self._color)
        noise_arr /= np.max(np.abs(noise_arr))  # normalize
        self._signal = np.vstack((noise_arr, noise_arr)).T * self._volume / 100

    # --------------------------------------------------------------------
    @staticmethod
    def _noise_psd(rng: np.random.Generator, N: int, color: str):
        """Compute the noise signal 1D array."""
        white = rng.standard_normal(size=N)
        dft = np.fft.rfft(white)
        S = _PSDS[color](np.fft.rfftfreq(N))
        # nornalize to preserve the energy from the white noise
        S /= np.sqrt(np.mean(S**2))
        return np.fft.irfft(dft * S)

    # --------------------------------------------------------------------
    @property
    def color(self) -> str:
        """The noise color."""
        logger.debug("'self._color' is set to %s.", self._color)
        return self._color

    @color.setter
    def color(self, color: str):
        logger.debug("Setting 'color' to %s.", color)
        _check_type(color, (str,), "color")
        _check_value(color, _PSDS)
        self._color = color
        self._set_signal()



"""Test _docs.py"""

import pytest

from .._docs import copy_doc, fill_doc


def test_fill_doc():
    """Test decorator to fill docstring."""
    # test filling docstring
    @fill_doc
    def foo(verbose):
        """My doc.

        Parameters
        ----------
        %(verbose)s
        """
        pass

    assert "verbose : int | str | bool | None" in foo.__doc__

    # test filling docstring with invalid key
    with pytest.raises(RuntimeError, match="Error documenting"):

        @fill_doc
        def foo(verbose):
            """My doc.

            Parameters
            ----------
            %(invalid_key)s
            """
            pass


def test_copy_doc():
    """Test decorator to copy docstring."""
    # test copy of docstring
    def foo(x, y):
        """
        My doc.
        """
        pass

    @copy_doc(foo)
    def foo2(x, y):
        pass

    assert "My doc." in foo2.__doc__

    # test copy of docstring from a function without docstring
    def foo(x, y):
        pass

    with pytest.raises(RuntimeError, match="The docstring from foo could not"):

        @copy_doc(foo)
        def foo2(x, y):
            pass



"""Module for auditory stimulus using sounddevice."""

from .am import SoundAM  # noqa: F401
from .base import BaseSound  # noqa: F401
from .noise import Noise  # noqa: F401
from .sound import Sound  # noqa: F401
from .tone import Tone  # noqa: F401



from copy import deepcopy
from typing import Optional, Tuple, Union

import cv2

from ..utils._checks import _check_type
from ..utils._docs import fill_doc
from .base import BaseFeedbackVisual


@fill_doc
class FillingBar(BaseFeedbackVisual):
    """Class to display a centered bar that can fill/unfill along a given axis.

    The filling process starts from the center of the bar and fills both sides
    simultaneously.

    Parameters
    ----------
    %(visual_window_name)s
    %(visual_window_size)s
    """

    def __init__(
        self,
        window_name: str = "Visual",
        window_size: Optional[Tuple[int, int]] = None,
    ):
        super().__init__(window_name, window_size)

    @fill_doc
    def putBar(
        self,
        length: int,
        width: int,
        margin: int,
        color: Union[str, Tuple[int, int, int]],
        fill_color: Union[str, Tuple[int, int, int]],
        fill_perc: float = 0,
        axis: Union[int, str] = 0,
    ) -> None:
        """Draw the bar on top of the current visual.

        Parameters
        ----------
        %(visual_length)s
        %(visual_width)s
        margin : int
            Margin in pixel between the filling bar and the containing bar.
            The containing bar (length x width) is set as:
                (length+margin, width+margin)
        color : str | tuple
            Color used to draw the bar background. %(visual_color)s
        color : str | tuple
            Color used to fill the bar. %(visual_color)s
        fill_perc : float
            Percentage between 0 and 1 of bar filling.
                - 0: not filled
                - 1: fully filled
            As the bar fills on both side simultaneously, the percentage filled
            is length//2 * fill_perc.
        axis : int | str
            Axis along which the bar is moving:
                - 0 | 'vertical' | 'v'      - vertical bar
                - 1 | 'horizontal' | 'h'    - horizontal bar
        """
        if self._backup_img is None:
            self._backup_img = deepcopy(self._img)
        else:
            self._reset()

        self._axis = BaseFeedbackVisual._check_axis(axis)
        self._length, margin = FillingBar._check_length_margin(
            length, margin, self._axis, self._window_size
        )
        self._width, margin = FillingBar._check_width_margin(
            width, margin, self._length, self._axis, self._window_size
        )
        self._margin = margin
        self._color = BaseFeedbackVisual._check_color(color)
        self._fill_color = BaseFeedbackVisual._check_color(fill_color)
        self._fill_perc = FillingBar._check_fill_perc(fill_perc)

        self._putBar()

    def _putBar(self) -> None:
        """Draw the bar rectangle and fill rectangle.

        - Axis = 1 - Horizontal bar
        P1 ---------------
        |                |
        --------------- P2

        - Axis = 0 - Vertical bar
        P1 ---
        |    |
        |    |
        |    |
        |    |
        |    |
        --- P2
        """
        # external rectangle to fill
        if self._axis == 0:
            xP1 = self.window_center[0] - self._width // 2 - self._margin
            yP1 = self.window_center[1] - self._length // 2 - self._margin
            xP2 = xP1 + self._width + 2 * self._margin
            yP2 = yP1 + self._length + 2 * self._margin
        elif self._axis == 1:
            xP1 = self.window_center[0] - self._length // 2 - self._margin
            yP1 = self.window_center[1] - self._width // 2 - self._margin
            xP2 = xP1 + self._length + 2 * self._margin
            yP2 = yP1 + self._width + 2 * self._margin

        cv2.rectangle(self._img, (xP1, yP1), (xP2, yP2), self._color, -1)

        # internal smaller rectangle filling the external rectangle
        fill_perc = int((self._length // 2) * self._fill_perc)
        if fill_perc != 0:
            if self._axis == 0:
                xP1 = self.window_center[0] - self._width // 2
                yP1 = self.window_center[1] - fill_perc
                xP2 = xP1 + self._width
                yP2 = yP1 + 2 * fill_perc
            elif self._axis == 1:
                xP1 = self.window_center[0] - fill_perc
                yP1 = self.window_center[1] - self._width // 2
                xP2 = xP1 + 2 * fill_perc
                yP2 = yP1 + self._width

            cv2.rectangle(
                self._img, (xP1, yP1), (xP2, yP2), self._fill_color, -1
            )

    # --------------------------------------------------------------------
    @staticmethod
    def _check_length_margin(
        length: int, margin: int, axis: int, window_size: Tuple[int, int]
    ) -> Tuple[int, int]:
        """Check that the length and margin are valid."""
        _check_type(length, ("int",), length)
        _check_type(margin, ("int",), margin)
        assert 0 < length
        assert 0 < margin
        assert margin < length
        assert length + margin <= window_size[(axis + 1) % 2]
        return length, margin

    @staticmethod
    def _check_width_margin(
        width: int,
        margin: int,
        length: int,
        axis: int,
        window_size: Tuple[int, int],
    ) -> Tuple[int, int]:
        """Check that the width is valid."""
        _check_type(width, ("int",), width)
        _check_type(margin, ("int",), margin)
        assert 0 < width
        assert width < length
        assert margin < width
        assert width + margin <= window_size[axis]
        return width, margin

    @staticmethod
    def _check_fill_perc(fill_perc: float) -> float:
        """Check that the fill length is a percentage between 0 and 1."""
        _check_type(fill_perc, ("numeric",), fill_perc)
        assert 0 <= fill_perc <= 1
        return fill_perc

    # --------------------------------------------------------------------
    @property
    def length(self) -> int:
        """Length of the bar in pixel."""
        return self._length

    @length.setter
    def length(self, length):
        self._length, _ = FillingBar._check_length_margin(
            length, self._margin, self._axis, self._window_size
        )
        self._reset()
        self._putBar()

    @property
    def width(self) -> int:
        """Width of the bar in pixel."""
        return self._width

    @width.setter
    def width(self, width):
        self._width, _ = FillingBar._check_width_margin(
            width, self._margin, self._length, self._axis, self._window_size
        )
        self._reset()
        self._putBar()

    @property
    def margin(self) -> int:
        """Margin in pixel between the bar and its filled content."""
        return self._margin

    @margin.setter
    def margin(self, margin):
        _, margin = FillingBar._check_length_margin(
            self._length, margin, self._axis, self._window_size
        )
        _, margin = FillingBar._check_width_margin(
            self._width, margin, self._length, self._axis, self._window_size
        )
        self._margin = margin
        self._reset()
        self._putBar()

    @property
    def color(self) -> Tuple[int, int, int]:
        """Color used for the bar background in BGR color space."""
        return self._color

    @color.setter
    def color(self, color):
        self._color = BaseFeedbackVisual._check_color(color)
        self._reset()
        self._putBar()

    @property
    def fill_color(self) -> Tuple[int, int, int]:
        """Color used to fill the bar in BGR color space."""
        return self._fill_color

    @fill_color.setter
    def fill_color(self, fill_color):
        self._fill_color = BaseFeedbackVisual._check_color(fill_color)
        self._reset()
        self._putBar()

    @property
    def fill_perc(self) -> float:
        """Length filled in percent between 0 and 1."""
        return self._fill_perc

    @fill_perc.setter
    def fill_perc(self, fill_perc):
        self._fill_perc = FillingBar._check_fill_perc(fill_perc)
        self._reset()
        self._putBar()

    @property
    def axis(self) -> int:
        """Axis on which the bar is moving.

        This property is a binary integer:
            - 0: Vertical bar filling along the vertical axis.
            - 1: Horizontal bar filling along the horizontal axis.
        """
        return self._axis

    @axis.setter
    def axis(self, axis):
        self._axis = BaseFeedbackVisual._check_axis(axis)
        self._reset()
        self._putBar()



"""Module for visual stimulus using opencv."""

from .base import BaseFeedbackVisual, BaseVisual  # noqa: F401
from .cross import Cross  # noqa: F401
from .filling_bar import FillingBar  # noqa: F401
from .moving_bar import MovingBar  # noqa: F401
from .text import Text  # noqa: F401



"""Test _logs.py"""

import logging
from typing import Optional, Union

import pytest

from .._logs import init_logger, logger, set_log_level, verbose

logger.propagate = True


def test_default_log_level(caplog):
    """Test the default log level."""
    init_logger()

    caplog.clear()
    logger.debug("101")
    assert "101" not in caplog.text

    caplog.clear()
    logger.info("101")
    assert "101" not in caplog.text

    caplog.clear()
    logger.warning("101")
    assert "101" in caplog.text

    caplog.clear()
    logger.error("101")
    assert "101" in caplog.text

    caplog.clear()
    logger.critical("101")
    assert "101" in caplog.text


@pytest.mark.parametrize(
    "level", ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL")
)
def test_logger(level, caplog):
    """Test basic logger functionalities."""
    level_functions = {
        "DEBUG": logger.debug,
        "INFO": logger.info,
        "WARNING": logger.warning,
        "ERROR": logger.error,
        "CRITICAL": logger.critical,
    }
    level_int = {
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
    }
    set_log_level(level)

    for level_, function in level_functions.items():
        caplog.clear()
        function("101")
        if level_int[level] <= level_int[level_]:
            assert "101" in caplog.text
        else:
            assert "101" not in caplog.text


def test_verbose(caplog):
    """Test verbose decorator."""

    @verbose
    def foo(verbose: Optional[Union[bool, str, int]] = None):
        logger.debug("101")

    set_log_level("DEBUG")
    caplog.clear()
    foo()
    assert "101" in caplog.text

    for level in (20, 25, 30, True, False, "WARNING", "ERROR"):
        caplog.clear()
        foo(verbose=level)
        assert "101" not in caplog.text

    caplog.clear()
    foo(verbose="DEBUG")
    assert "101" in caplog.text



from typing import Optional, Tuple, Union

import cv2

from ..utils._checks import _check_type
from ..utils._docs import fill_doc
from .base import BaseVisual


@fill_doc
class Text(BaseVisual):
    """
    Class to display a text.

    Parameters
    ----------
    %(visual_window_name)s
    %(visual_window_size)s
    """

    def __init__(
        self,
        window_name: str = "Visual",
        window_size: Optional[Tuple[int, int]] = None,
    ):
        super().__init__(window_name, window_size)

    @fill_doc
    def putText(
        self,
        text: str,
        fontFace: int = cv2.FONT_HERSHEY_DUPLEX,
        fontScale: int = 2,
        color: Union[str, Tuple[int, int, int]] = "white",
        thickness: int = 2,
        lineType: int = cv2.LINE_AA,
        position: Union[str, Tuple[int, int]] = "centered",
    ) -> None:
        """Add text to the visual.

        Parameters
        ----------
        text : str
            Text to display.
        fontFace : int (cv2.FONT)
            Font to use to write the text.
        fontScale : int
            Font scale factor multiplied by the font-specific base size.
        color : str | tuple
            Color used to write the text. %(visual_color)s
        thickness : int
            Text line thickness in pixel.
        lineType : int (cv2.LINE)
            Type of line to use.
        position : str | tuple
            Position of the bottom left corner of the text. See notes for
            additional information.

        Notes
        -----
        %(visual_position)s
        """
        _check_type(text, (str,), "text")
        if len(text.strip()) == 0:
            return None
        _check_type(fontFace, ("int",), "fontFace")
        _check_type(fontScale, ("int",), "fontScale")
        _check_type(thickness, ("int",), "thickness")
        _check_type(lineType, ("int",), "lineType")
        textWidth, textHeight = cv2.getTextSize(
            text, fontFace, fontScale, thickness
        )[0]
        position = Text._check_position(
            position,
            textWidth,
            textHeight,
            self._window_size,
            self._window_center,
        )
        color = BaseVisual._check_color(color)

        cv2.putText(
            self._img,
            text,
            position,
            fontFace,
            fontScale,
            color,
            thickness=thickness,
            lineType=lineType,
        )

    # --------------------------------------------------------------------
    @staticmethod
    def _check_position(
        position: Union[str, Tuple[int, int]],
        textWidth,
        textHeight,
        window_size: Tuple[int, int],
        window_center: Tuple[int, int],
    ) -> Tuple[int, int]:
        """Check that the text position is coherent with the window size.

        The position of the text is given as the bottom left corner of the
        text-box.
        """
        _check_type(position, (str, tuple), "position")
        if isinstance(position, str):
            position = position.lower().strip()
            assert position in ["centered", "center"]
            position = (
                window_center[0] - textWidth // 2,
                window_center[1] + textHeight // 2,
            )
        for pos in position:
            _check_type(pos, ("int",))
        assert len(position) == 2
        assert 0 <= position[0]
        assert position[0] + textWidth <= window_size[0]
        assert 0 <= position[1] - textHeight
        assert position[1] <= window_size[1]
        return position



from itertools import product

import numpy as np
import pytest

from .. import Cross
from .test_base import _test_base


@pytest.mark.parametrize(
    "length, thickness, color, position",
    product(
        (100, 20),
        (5, 10),
        ("white", "teal", (101, 101, 101)),
        ("center", (100, 400)),
    ),
)
def test_cross(length, thickness, color, position):
    """Test a cross visual."""
    visual = Cross(window_name="test", window_size=(500, 500))
    assert visual.window_name == "test"
    assert visual.window_size == (500, 500)
    assert visual._window_center == (250, 250)
    assert np.count_nonzero(visual.img) == 0
    visual.putCross(length, thickness, color, position)


def test_base():
    """Test base functionalities with a cross visual."""
    _test_base(Cross)


def test_invalid_arguments():
    """Test invalid arguments."""
    visual = Cross(window_name="test", window_size=(500, 500))

    with pytest.raises(TypeError, match="must be an instance of"):
        visual.putCross(length="long", thickness=10, color="white")
    with pytest.raises(TypeError, match="must be an instance of"):
        visual.putCross(length=100, thickness="large", color="white")

    with pytest.raises(AssertionError):
        visual.putCross(length=-100, thickness=10, color="white")
    with pytest.raises(AssertionError):
        visual.putCross(length=100, thickness=-10, color="white")
    with pytest.raises(AssertionError):
        visual.putCross(length=10000, thickness=20, color="white")
    with pytest.raises(AssertionError):
        visual.putCross(length=100, thickness=200, color="white")
    with pytest.raises(AssertionError):
        visual.putCross(
            length=100, thickness=10, color="white", position=(10, 10, 10)
        )
    with pytest.raises(AssertionError):
        visual.putCross(
            length=100, thickness=10, color="white", position=(-10, 10)
        )
    with pytest.raises(AssertionError):
        visual.putCross(
            length=100, thickness=10, color="white", position=(1, 1)
        )
    with pytest.raises(AssertionError):
        visual.putCross(
            length=100, thickness=10, color="white", position=(499, 250)
        )



import argparse

from .. import set_log_level
from ..visuals import Cross
from . import helpdict


def run():
    """Entry-point to display a fixation cross via CLI."""
    parser = argparse.ArgumentParser(
        prog="Stimuli", description="Display a fixation cross"
    )
    parser.add_argument(
        "--name",
        help="name of the visual window",
        type=str,
        metavar=str,
        default="FixationCross",
    )
    parser.add_argument(
        "--winsize",
        help="size of the visual window",
        type=int,
        metavar=int,
        nargs=2,
        default=None,
    )
    parser.add_argument(
        "--bgcolor",
        help="background BGR color as 3 integers in the range [0, 255]",
        type=int,
        metavar=int,
        nargs=3,
        default=(0, 0, 0),
    )
    parser.add_argument(
        "--length",
        help="length of the fixation cross [px]",
        type=int,
        metavar=int,
        default=200,
    )
    parser.add_argument(
        "--thickness",
        help="thickness of the fixation cross [px]",
        type=int,
        metavar=int,
        default=30,
    )
    parser.add_argument(
        "--color",
        help="BGR color of the cross as 3 integers in the range [0, 255]",
        type=int,
        metavar=int,
        nargs=3,
        default=(210, 210, 210),
    )
    parser.add_argument(
        "--position",
        help="(x, y) position of the fixation cross [px]",
        type=str,
        metavar=int,
        nargs=2,
        default="center",
    )
    parser.add_argument(
        "--loglevel",
        type=str,
        metavar="str",
        help=helpdict["loglevel"],
        default="warning",
    )
    args = parser.parse_args()
    set_log_level(args.loglevel.upper().strip())

    winsize = None if args.winsize is None else tuple(args.winsize)
    position = (
        args.position
        if args.position in ("center", "centered")
        else tuple([int(elt) for elt in args.position])
    )

    visual = Cross(args.name, winsize)
    visual.draw_background(tuple(args.bgcolor))
    visual.putCross(args.length, args.thickness, tuple(args.color), position)
    visual.show(0)



from itertools import product

import numpy as np
import pytest

from ... import logger
from .. import Noise, Sound, SoundAM, Tone

logger.propagate = True

SoundClasses = (SoundAM, Tone, Noise)


@pytest.mark.parametrize("SoundClass", SoundClasses)
def test_io(tmp_path, SoundClass):
    """Test save/load for all sound classes."""
    fname = tmp_path / f"sound-{SoundClass.__name__}-1.wav"
    sound = SoundClass(volume=100)
    sound.save(fname, overwrite=False)
    sound_loaded = Sound(fname)
    assert sound_loaded.volume == 100
    assert np.allclose(sound.signal, sound_loaded.signal)

    fname = tmp_path / f"sound-{SoundClass.__name__}-2.wav"
    sound = SoundClass(volume=10)
    sound.save(fname, overwrite=False)
    sound_loaded = Sound(fname)
    assert sound_loaded.volume == 100
    sound_loaded.volume = 10
    assert np.allclose(sound.signal, sound_loaded.signal)

    fname = tmp_path / f"sound-{SoundClass.__name__}-3.wav"
    sound = SoundClass(volume=(50, 25))
    sound.save(fname, overwrite=False)
    sound_loaded = Sound(fname)
    assert np.allclose(sound_loaded.volume, (100, 50))
    sound_loaded.volume = (50, 25)
    assert np.allclose(sound.signal, sound_loaded.signal)


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@pytest.mark.parametrize("SoundClass", SoundClasses)
def test_io_with_muted_channel(tmp_path, SoundClass):
    """Test save/load for all sound classes with a muted channel."""
    fname = tmp_path / f"sound-{SoundClass.__name__}-1.wav"
    sound = SoundClass(volume=(50, 0))
    sound.save(fname, overwrite=True)
    sound_loaded = Sound(fname)
    assert np.allclose(sound_loaded.volume, (100, 0))
    sound_loaded.volume = (50, 0)
    assert np.allclose(sound.signal, sound_loaded.signal)


@pytest.mark.parametrize("SoundClass", SoundClasses)
def test_overwrite(tmp_path, SoundClass):
    """Test the overwrite argument."""
    fname = tmp_path / f"sound-{SoundClass.__name__}-1.wav"
    sound = SoundClass(volume=100)
    sound.save(fname, overwrite=True)
    sound2 = SoundClass(volume=100, duration=2)
    with pytest.raises(RuntimeError, match="already exist"):
        sound2.save(fname, overwrite=False)
    sound2.save(fname, overwrite=True)
    sound_loaded = Sound(fname)
    assert sound_loaded.volume == 100
    assert sound.signal.shape != sound_loaded.signal.shape
    assert np.allclose(sound2.signal, sound_loaded.signal)


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@pytest.mark.parametrize("SoundClass", SoundClasses)
def test_reset(tmp_path, SoundClass):
    """Test the reset method."""
    # with a change of volume
    fname = tmp_path / f"sound-{SoundClass.__name__}-1.wav"
    sound = SoundClass(volume=100)
    sound.save(fname, overwrite=True)
    sound_loaded = Sound(fname)
    assert sound_loaded.volume == 100
    sound_loaded.volume = 5
    assert sound_loaded.volume == 5
    assert not np.allclose(sound_loaded.signal, sound.signal)
    sound_loaded.reset()
    assert np.allclose(sound_loaded.signal, sound.signal)
    assert sound_loaded.volume == 100

    # with a channel muted
    fname = tmp_path / f"sound-{SoundClass.__name__}-2.wav"
    sound = SoundClass(volume=(100, 0))
    sound.save(fname, overwrite=True)
    sound_loaded = Sound(fname)
    assert np.allclose(sound_loaded.volume, (100, 0))
    sound_loaded.volume = (5, 0)
    assert np.allclose(sound_loaded.volume, (5, 0))
    assert not np.allclose(sound_loaded.signal, sound.signal)
    sound_loaded.reset()
    assert np.allclose(sound_loaded.signal, sound.signal)
    assert np.allclose(sound_loaded.volume, (100, 0))

    # with a crop
    fname = tmp_path / f"sound-{SoundClass.__name__}-3.wav"
    sound = SoundClass(volume=100, duration=10)
    sound.save(fname, overwrite=True)
    sound_loaded = Sound(fname)
    assert sound_loaded.volume == 100
    sound_loaded.crop(tmin=1, tmax=5)
    assert sound_loaded.signal.shape != sound.signal.shape
    sound_loaded.reset()
    assert sound_loaded.signal.shape == sound.signal.shape


@pytest.mark.parametrize(
    "SoundClass, tmin, tmax", product(SoundClasses, (1, None), (5, None))
)
def test_crop(tmp_path, SoundClass, tmin, tmax):
    """Test the crop method."""
    fname = tmp_path / f"sound-{SoundClass.__name__}-1.wav"
    sound = SoundClass(volume=100, duration=10)
    sound.save(fname, overwrite=True)
    sound_loaded = Sound(fname)
    assert sound_loaded.duration == sound.duration
    sound_loaded.crop(tmin=tmin, tmax=tmax)
    tmin = sound_loaded.times[0] if tmin is None else tmin
    tmax = sound_loaded.times[-1] if tmax is None else tmax
    assert np.isclose(
        sound_loaded.duration, tmax - tmin, atol=2 / sound_loaded.sample_rate
    )
    assert np.isclose(
        sound_loaded.times[-1] - sound_loaded.times[0],
        sound_loaded.duration,
        atol=2 / sound_loaded.sample_rate,
    )
    assert np.isclose(
        sound_loaded._times[-1] - sound_loaded._times[0],
        sound_loaded.duration,
        atol=2 / sound_loaded.sample_rate,
    )
    assert sound_loaded.times.size == sound_loaded.signal.shape[0]


@pytest.mark.parametrize("SoundClass", SoundClasses)
def test_successive_crop(tmp_path, SoundClass):
    """Test successive crops."""
    fname = tmp_path / f"sound-{SoundClass.__name__}-1.wav"
    sound = SoundClass(volume=100, duration=10)
    sound.save(fname, overwrite=True)
    sound_loaded = Sound(fname)
    assert sound_loaded.tmin == 0
    assert sound_loaded.tmax == 10
    sound_loaded.crop(tmin=1, tmax=8)
    assert np.isclose(sound_loaded.tmin, 1, atol=2 / sound_loaded.sample_rate)
    assert np.isclose(sound_loaded.tmax, 8, atol=2 / sound_loaded.sample_rate)
    sound_loaded.crop(tmin=1, tmax=None)
    assert np.isclose(sound_loaded.tmin, 1, atol=2 / sound_loaded.sample_rate)
    assert np.isclose(sound_loaded.tmax, 10, atol=2 / sound_loaded.sample_rate)
    sound_loaded.crop(tmin=1, tmax=5)
    assert np.isclose(sound_loaded.tmin, 1, atol=2 / sound_loaded.sample_rate)
    assert np.isclose(sound_loaded.tmax, 5, atol=2 / sound_loaded.sample_rate)


@pytest.mark.parametrize("SoundClass", SoundClasses)
def test_sample_rate(tmp_path, caplog, SoundClass):
    """Test the sample-rate property."""
    fname = tmp_path / f"sound-{SoundClass.__name__}-1.wav"
    sound = SoundClass(volume=100)
    sound.save(fname, overwrite=True)
    sound_loaded = Sound(fname)
    assert sound_loaded.sample_rate == sound.sample_rate
    caplog.clear()
    sound_loaded.sample_rate = 101
    assert "The sampling rate property" in caplog.text
    assert sound_loaded.sample_rate == sound.sample_rate


@pytest.mark.parametrize("SoundClass", SoundClasses)
def test_duration(tmp_path, caplog, SoundClass):
    """Test the duration property."""
    fname = tmp_path / f"sound-{SoundClass.__name__}-1.wav"
    sound = SoundClass(volume=100)
    sound.save(fname, overwrite=True)
    sound_loaded = Sound(fname)
    assert sound_loaded.duration == sound.duration
    caplog.clear()
    sound_loaded.duration = 101
    assert "The duration property" in caplog.text
    assert sound_loaded.duration == sound.duration



from typing import Optional, Tuple, Union

import cv2

from ..utils._checks import _check_type
from ..utils._docs import fill_doc
from .base import BaseVisual


@fill_doc
class Cross(BaseVisual):
    """Class to display a cross, e.g. a fixation cross.

    Parameters
    ----------
    %(visual_window_name)s
    %(visual_window_size)s
    """

    def __init__(
        self,
        window_name: str = "Visual",
        window_size: Optional[Tuple[int, int]] = None,
    ):
        super().__init__(window_name, window_size)

    @fill_doc
    def putCross(
        self,
        length: int,
        thickness: int,
        color: Union[str, Tuple[int, int, int]],
        position: Union[str, Tuple[int, int]] = "centered",
    ) -> None:
        """Draw a cross composed of 2 rectangles.

        The rectangles are defined by length and thickness.
        The rectangles are positioned to form a cross.

        - Horizontal rectangle
        P1 --------------
        |                |
         -------------- P2

        - Vertical rectangle
        P1 --
        |    |
        |    |
        |    |
        |    |
        |    |
         -- P2

        Parameters
        ----------
        length : int
            Number of pixels used to draw the length of the cross.
        thickness : int
            Number of pixels used to draw the thickness of the cross.
        color : str | tuple
            Color used to fill the cross. %(visual_color)s
        position : str | tuple
            Position of the center of the cross. See notes for additional
            information.

        Notes
        -----
        %(visual_position)s
        """
        length = Cross._check_length(length, self._window_size)
        thickness = Cross._check_thickness(thickness, length)
        color = BaseVisual._check_color(color)
        position = Cross._check_position(
            position, length, self._window_size, self._window_center
        )

        # Horizontal rectangle
        xP1 = position[0] - length // 2
        yP1 = position[1] - thickness // 2
        xP2 = xP1 + length
        yP2 = yP1 + thickness
        cv2.rectangle(self._img, (xP1, yP1), (xP2, yP2), color, -1)

        # Vertical rectangle
        xP1 = position[0] - thickness // 2
        yP1 = position[1] - length // 2
        xP2 = xP1 + thickness
        yP2 = yP1 + length
        cv2.rectangle(self._img, (xP1, yP1), (xP2, yP2), color, -1)

    # --------------------------------------------------------------------
    @staticmethod
    def _check_length(length: int, window_size: Tuple[int, int]) -> int:
        """Check that the length is valid."""
        _check_type(length, ("int",), "length")
        assert 0 < length
        assert all(length <= size for size in window_size)
        return length

    @staticmethod
    def _check_thickness(thickness: int, length: int) -> int:
        """Check that the thickness is valid."""
        _check_type(thickness, ("int",), "thickness")
        assert 0 < thickness
        assert thickness < length
        return thickness

    @staticmethod
    def _check_position(
        position: Union[str, Tuple[int, int]],
        length: int,
        window_size: Tuple[int, int],
        window_center: Tuple[int, int],
    ) -> Tuple[int, int]:
        """Check that the cross position is coherent with the window size.

        The position of the cross is given as the center of the cross.
        """
        _check_type(position, (str, tuple), "position")
        if isinstance(position, str):
            position = position.lower().strip()
            assert position in ["centered", "center"]
            position = window_center
        for pos in position:
            _check_type(pos, ("int",))
        assert len(position) == 2
        assert 0 <= position[0] - length // 2
        assert position[0] - length // 2 + length <= window_size[0]
        assert 0 <= position[1] - length // 2
        assert position[1] - length // 2 + length <= window_size[1]
        return position



"""Sound loaded from a file."""

from pathlib import Path
from typing import Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray
from scipy.io import wavfile

from .. import logger
from ..utils._checks import _check_type
from ..utils._docs import copy_doc
from .base import BaseSound


class Sound(BaseSound):
    """Auditory stimulus loaded from a file.

    Parameters
    ----------
    fname : str | Path
        Path to the supported audio file to load.
    """

    def __init__(self, fname: Union[str, Path]):
        self._fname = BaseSound._check_file(fname, must_exists=True)

        sample_rate, original_signal = wavfile.read(self._fname)
        self._original_signal, volume = Sound._check_signal(original_signal)
        self._original_duration = self._original_signal.shape[0] / sample_rate

        self._tmin = None  # idx
        self._tmax = None  # idx
        super().__init__(volume, sample_rate, self._original_duration)
        self._original_times = self._times.copy()

    @copy_doc(BaseSound._set_signal)
    def _set_signal(self) -> None:
        assert self._original_signal.ndim == 2
        tmax = None if self._tmax is None else self._tmax + 1  # +1 for slice
        slc = (slice(self._tmin, tmax), slice(None))
        self._signal = self._original_signal[slc] * self._volume / 100

    def crop(
        self, tmin: Optional[float] = None, tmax: Optional[float] = None
    ) -> None:
        """Crop the sound between tmin and tmax.

        Parameters
        ----------
        tmin : float | None
            Left-edge of the crop. If None, the beginning of the sound.
        tmax : float | None
            Right-edge of the crop. If None, the end of the sound.

        Notes
        -----
        The time-based selection selects the samples in the closed interval
        [tmin, tmax].
        """
        logger.debug("Cropping the signal between %s and %s.", tmin, tmax)
        self._tmin, self._tmax = Sound._check_tmin_tmax(
            tmin, tmax, self._original_times
        )
        logger.debug(
            "'tmin' corresponds to the idx %i and 'tmax' corresponds "
            "to the idx %i.",
            self._tmin,
            self._tmax,
        )
        self._duration = (
            self._original_times[self._tmax] - self._original_times[self._tmin]
        )
        # tmax + 1 for slice
        self._times = self._original_times[self._tmin : self._tmax + 1]
        self._set_signal()

    def reset(self) -> None:
        """Reset the signal to the original loaded signal."""
        logger.debug("Resetting the loaded sound.")
        self._tmin = None
        self._tmax = None
        self._duration = self._original_duration
        self._volume = np.max(np.abs(self._original_signal), axis=0) * 100
        self._set_times()
        self._set_signal()

    # --------------------------------------------------------------------
    @staticmethod
    def _check_signal(
        signal: NDArray[float],
    ) -> Tuple[NDArray[float], Tuple[float, float]]:
        """Check that the sound is either mono or stereo.

        Parameters
        ----------
        signal : array
            Raw signal loaded from the file.

        Returns
        -------
        signal : array
            Signal that has been normalized channel-wise.
        volume : tuple
            A 2-float tuple representing the (L, R) volume before channel-wise
            normalization. The volume is normalized between 0 and 100.

        Notes
        -----
        The volume is retrieved after normalizing the signal with the max of
        both channels, thus preserving the difference between both channels.
        The returned signal, stored in _original_signal, is normalized channel-
        wise. Thus, it does not correspond to the returned volume.

        However, the set signal, stored in _signal, is scaled by the volume in
        the _set_signal method.
        """
        assert signal.ndim in (1, 2)
        if signal.ndim == 2:
            assert signal.shape[1] in (1, 2)
            if signal.shape[1] == 1:
                signal = signal[:, 0]
        if signal.ndim == 1:
            signal = np.vstack((signal, signal)).T
        # normalize to retrieve the volume of boh channels
        max_ = np.max(np.abs(signal))
        if max_ == 0:
            logger.warning("The loaded sound has 2 empty channels.")
            return signal(0, 0)
        signal /= max_
        volume = tuple(np.max(np.abs(signal), axis=0) * 100)
        assert len(volume) == 2  # sanity-check
        assert any(elt == 100 for elt in volume)  # sanity-check
        # normalize both channels
        if all(elt != 0 for elt in volume):
            signal /= np.max(np.abs(signal), axis=0)
            np.nan_to_num(signal, copy=False, nan=0.0)  # sanity-check
        return signal, volume

    @staticmethod
    def _check_tmin_tmax(
        tmin: Optional[float], tmax: Optional[float], times: NDArray[float]
    ) -> Tuple[int, int]:
        """Check tmin/tmax and convert to idx."""
        _check_type(tmin, ("numeric", None), "tmin")
        _check_type(tmax, ("numeric", None), "tmax")
        tmin = 0 if tmin is None else tmin
        tmin = tmin if np.isfinite(tmin) else 0
        tmax = times[-1] if tmax is None else tmax
        tmax = tmax if np.isfinite(tmax) else times[-1]
        assert 0 <= tmin
        assert tmax <= times[-1]
        idx = np.where((tmin <= times) & (times <= tmax))[0]
        return idx[0], idx[-1]

    # --------------------------------------------------------------------
    @BaseSound.sample_rate.setter
    def sample_rate(self, sample_rate: int):
        logger.warning(
            "The sampling rate property of a loaded sound can not be "
            "changed. Skipping."
        )

    @BaseSound.duration.setter
    def duration(self, duration: float):
        logger.warning(
            "The duration property of a loaded sound can not be "
            "changed. Use the method .crop(tmin, tmax) to trim "
            "a loaded sound. Skipping."
        )

    @property
    def fname(self) -> Path:
        """The sound's original file name."""
        logger.debug("'self._fname' is set to %s.", self._fname)
        return self._fname

    @property
    def tmin(self) -> float:
        """Left-edge of the signal crop [seconds]."""
        logger.debug("'self._tmin' is set to %s [AU].", self._tmin)
        if self._tmin is None:
            return self._times[0]
        return self._original_times[self._tmin]

    @property
    def tmax(self) -> float:
        """Right-edge of the signal crop [seconds]."""
        logger.debug("'self._tmax' is set to %s [AU].", self._tmax)
        if self._tmax is None:
            return self._times[-1]
        return self._original_times[self._tmax]



"""Pure tone sound."""

from typing import Tuple, Union

import numpy as np

from .. import logger
from ..utils._checks import _check_type
from ..utils._docs import copy_doc, fill_doc
from .base import BaseSound


@fill_doc
class Tone(BaseSound):
    """Pure tone stimulus at the frequency f (Hz).

    The equation is sin(2*pi*f*time).
    Example: A 440 - La 440 - Tone(f=440)

    Parameters
    ----------
    %(audio_volume)s
    %(audio_sample_rate)s
    %(audio_duration)s
    frequency : float
        Pure tone frequency. The default is 440 Hz (La - A440).
    """

    def __init__(
        self,
        volume: Union[float, Tuple[float, float]],
        sample_rate: int = 44100,
        duration: float = 1,
        frequency: float = 440,
    ):
        self._frequency = Tone._check_frequency(frequency)
        self.name = "tone"
        super().__init__(volume, sample_rate, duration)

    @copy_doc(BaseSound._set_signal)
    def _set_signal(self) -> None:
        tone_arr = np.sin(2 * np.pi * self._frequency * self._times)
        tone_arr /= np.max(np.abs(tone_arr))  # normalize
        self._signal = np.vstack((tone_arr, tone_arr)).T * self._volume / 100

    # --------------------------------------------------------------------
    @staticmethod
    def _check_frequency(frequency: float) -> float:
        """Check if the frequency is positive."""
        _check_type(frequency, ("numeric",), item_name="frequency")
        assert 0 < frequency
        return frequency

    # --------------------------------------------------------------------
    @property
    def frequency(self) -> float:
        """Sound's pure tone frequency [Hz]."""
        logger.debug("'self._frequency' is set to %.2f [Hz].", self._frequency)
        return self._frequency

    @frequency.setter
    def frequency(self, frequency: float):
        logger.debug("Setting 'frequency' to %.2f [Hz].", frequency)
        self._frequency = Tone._check_frequency(frequency)
        self._set_signal()



"""Utility functions for checking types and values. Inspired from MNE."""

import logging
import operator
import os
from pathlib import Path
from typing import Any, Optional, Union

import numpy as np


def _ensure_int(item: Any, item_name: Optional[str] = None) -> int:
    """Ensure a variable is an integer.

    Parameters
    ----------
    item : Any
        Item to check.
    item_name : str | None
        Name of the item to show inside the error message.

    Raises
    ------
    TypeError
        When the type of the item is not int.
    """
    # This is preferred over numbers.Integral, see:
    # https://github.com/scipy/scipy/pull/7351#issuecomment-299713159
    try:
        # someone passing True/False is much more likely to be an error than
        # intentional usage
        if isinstance(item, bool):
            raise TypeError
        item = int(operator.index(item))
    except TypeError:
        item_name = "Item" if item_name is None else "'%s'" % item_name
        raise TypeError(
            "%s must be an int, got %s instead." % (item_name, type(item))
        )

    return item


class _IntLike:
    @classmethod
    def __instancecheck__(cls, other: Any) -> bool:
        try:
            _ensure_int(other)
        except TypeError:
            return False
        else:
            return True


class _Callable:
    @classmethod
    def __instancecheck__(cls, other: Any) -> bool:
        return callable(other)


_types = {
    "numeric": (np.floating, float, _IntLike()),
    "path-like": (str, Path, os.PathLike),
    "int": (_IntLike(),),
    "callable": (_Callable(),),
}


def _check_type(
    item: Any, types: tuple, item_name: Optional[str] = None
) -> Any:
    """Check that item is an instance of types.

    Parameters
    ----------
    item : object
        Item to check.
    types : tuple of types | tuple of str
        Types to be checked against.
        If str, must be one of:
            ('int', 'str', 'numeric', 'path-like', 'callable')
    item_name : str | None
        Name of the item to show inside the error message.

    Raises
    ------
    TypeError
        When the type of the item is not one of the valid options.
    """
    check_types = sum(
        (
            (type(None),)
            if type_ is None
            else (type_,)
            if not isinstance(type_, str)
            else _types[type_]
            for type_ in types
        ),
        (),
    )

    if not isinstance(item, check_types):
        type_name = [
            "None"
            if cls_ is None
            else cls_.__name__
            if not isinstance(cls_, str)
            else cls_
            for cls_ in types
        ]
        if len(type_name) == 1:
            type_name = type_name[0]
        elif len(type_name) == 2:
            type_name = " or ".join(type_name)
        else:
            type_name[-1] = "or " + type_name[-1]
            type_name = ", ".join(type_name)
        item_name = "Item" if item_name is None else "'%s'" % item_name
        raise TypeError(
            f"{item_name} must be an instance of {type_name}, "
            f"got {type(item)} instead."
        )

    return item


def _check_value(
    item: Any,
    allowed_values: tuple,
    item_name: Optional[str] = None,
    extra: Optional[str] = None,
) -> Any:
    """Check the value of a parameter against a list of valid options.

    Parameters
    ----------
    item : object
        Item to check.
    allowed_values : tuple of objects
        Allowed values to be checked against.
    item_name : str | None
        Name of the item to show inside the error message.
    extra : str | None
        Extra string to append to the invalid value sentence, e.g.
        "when using ico mode".

    Raises
    ------
    ValueError
        When the value of the item is not one of the valid options.
    """
    if item not in allowed_values:
        item_name = "" if item_name is None else " '%s'" % item_name
        extra = "" if extra is None else " " + extra
        msg = (
            "Invalid value for the{item_name} parameter{extra}. "
            "{options}, but got {item!r} instead."
        )
        allowed_values = tuple(allowed_values)  # e.g., if a dict was given
        if len(allowed_values) == 1:
            options = "The only allowed value is %s" % repr(allowed_values[0])
        elif len(allowed_values) == 2:
            options = "Allowed values are %s and %s" % (
                repr(allowed_values[0]),
                repr(allowed_values[1]),
            )
        else:
            options = "Allowed values are "
            options += ", ".join([f"{repr(v)}" for v in allowed_values[:-1]])
            options += f", and {repr(allowed_values[-1])}"
        raise ValueError(
            msg.format(
                item_name=item_name, extra=extra, options=options, item=item
            )
        )

    return item


def _check_verbose(verbose: Union[bool, str, int, None]) -> int:
    """Check that the value of verbose is valid.

    Parameters
    ----------
    verbose : bool | str | int | None
        The verbosity level.

    Returns
    -------
    verbose : int
        The verbosity level as an integer.
    """
    logging_types = dict(
        DEBUG=logging.DEBUG,
        INFO=logging.INFO,
        WARNING=logging.WARNING,
        ERROR=logging.ERROR,
        CRITICAL=logging.CRITICAL,
    )

    _check_type(verbose, (bool, str, "int", None), item_name="verbose")

    if verbose is None:
        verbose = logging.WARNING
    elif isinstance(verbose, str):
        verbose = verbose.upper()
        _check_value(verbose, logging_types, item_name="verbose")
        verbose = logging_types[verbose]
    elif isinstance(verbose, bool):
        if verbose:
            verbose = logging.INFO
        else:
            verbose = logging.WARNING
    elif isinstance(verbose, int):
        if verbose <= 0:
            raise ValueError(
                "Argument 'verbose' can not be a negative integer, "
                f"{verbose} is invalid."
            )

    return verbose



from itertools import product

import cv2
import numpy as np
import pytest

from .. import Text
from .test_base import _test_base


@pytest.mark.parametrize(
    "fontFace, fontScale, color, thickness, lineType, position",
    product(
        (cv2.FONT_HERSHEY_COMPLEX, cv2.FONT_HERSHEY_SIMPLEX),
        (2, 3),
        ("white", "teal", (101, 101, 101)),
        (2, 4),
        (cv2.LINE_AA, cv2.LINE_4, cv2.LINE_8),
        ("center", (100, 400)),
    ),
)
def test_text(fontFace, fontScale, color, thickness, lineType, position):
    """Test a text visual."""
    visual = Text(window_name="test", window_size=(500, 500))
    assert visual.window_name == "test"
    assert visual.window_size == (500, 500)
    assert visual._window_center == (250, 250)
    assert np.count_nonzero(visual.img) == 0
    visual.putText(
        "101", fontFace, fontScale, color, thickness, lineType, position
    )


def test_base():
    """Test base functionalities with a text visual."""
    _test_base(Text)


def test_invalid_arguments():
    """Test invalid arguments."""
    visual = Text(window_name="test", window_size=(500, 500))

    with pytest.raises(TypeError, match="must be an instance of"):
        visual.putText(101)
    with pytest.raises(TypeError, match="must be an instance of"):
        visual.putText("101", fontFace="arial")
    with pytest.raises(TypeError, match="must be an instance of"):
        visual.putText("101", fontScale="large")
    with pytest.raises(TypeError, match="must be an instance of"):
        visual.putText("101", color=101)
    with pytest.raises(TypeError, match="must be an instance of"):
        visual.putText("101", thickness="large")
    with pytest.raises(TypeError, match="must be an instance of"):
        visual.putText("101", lineType="large")

    with pytest.raises(AssertionError):
        visual.putText("101", position=(10, 10, 10))
    with pytest.raises(AssertionError):
        visual.putText("101", position=(-10, 10))
    with pytest.raises(AssertionError):
        visual.putText("101", position=(1, 1))
    with pytest.raises(AssertionError):
        visual.putText("101", position=(499, 400))



"""Fill docstrings to avoid redundant docstrings in multiple files.

Inspired from mne: https://mne.tools/stable/index.html
Inspired from mne.utils.docs.py by Eric Larson <larson.eric.d@gmail.com>
"""

import sys
from typing import Callable, List

# ------------------------- Documentation dictionary -------------------------
docdict = dict()

# ---------------------------------- verbose ---------------------------------
docdict[
    "verbose"
] = """
verbose : int | str | bool | None
    Sets the verbosity level. The verbosity increases gradually between
    "CRITICAL", "ERROR", "WARNING", "INFO" and "DEBUG".
    If None is provided, the verbosity is set to "WARNING".
    If a bool is provided, the verbosity is set to "WARNING" for False and to
    "INFO" for True."""

# ----------------------------------- audio ----------------------------------
docdict[
    "audio_volume"
] = """
volume : float | tuple
    If an int or a float is provided, the sound will use only one channel
    (mono). If a 2-length tuple is provided, the sound will use 2
    channels (stereo). The volume of each channel is given between 0 and 100.
    For stereo, the volume is given as (L, R)."""
docdict[
    "audio_sample_rate"
] = """
sample_rate : int
    Sampling frequency of the sound. The default is 44100 kHz."""
docdict[
    "audio_duration"
] = """
duration : float
    Duration of the sound. The default is 1 second."""

# ----------------------------------- visual ----------------------------------
docdict[
    "visual_window_name"
] = """
window_name : str
    Name of the window in which the visual is displayed."""
docdict[
    "visual_window_size"
] = """
window_size : tuple | None
    Either None to automatically select a window size based on the
    available monitors, or a 2-length of positive integer sequence as
    (width, height)."""
docdict[
    "visual_color"
] = """The color is provided as a matplotlib string or a
(B, G, R) tuple of int8 set between 0 and 255."""
docdict[
    "visual_position"
] = """
The position of the object can be either defined as the string 'center' or
'centered' to position the object in the center of the window; or as a 2-length
tuple of positive integer. The position is defined in opencv coordinates, with
(0, 0) being the top left corner of the window."""
docdict[
    "visual_length"
] = """
length : int
    Number of pixels used to draw the length of the bar."""
docdict[
    "visual_width"
] = """
width : int
    Number of pixels used to draw the width of the bar."""

# ------------------------- Documentation functions --------------------------
docdict_indented = dict()


def fill_doc(f: Callable) -> Callable:
    """Fill a docstring with docdict entries.

    Parameters
    ----------
    f : callable
        The function to fill the docstring of (modified in place).

    Returns
    -------
    f : callable
        The function, potentially with an updated __doc__.
    """
    docstring = f.__doc__
    if not docstring:
        return f

    lines = docstring.splitlines()
    indent_count = _indentcount_lines(lines)

    try:
        indented = docdict_indented[indent_count]
    except KeyError:
        indent = " " * indent_count
        docdict_indented[indent_count] = indented = dict()

        for name, docstr in docdict.items():
            lines = [
                indent + line if k != 0 else line
                for k, line in enumerate(docstr.strip().splitlines())
            ]
            indented[name] = "\n".join(lines)

    try:
        f.__doc__ = docstring % indented
    except (TypeError, ValueError, KeyError) as exp:
        funcname = f.__name__
        funcname = docstring.split("\n")[0] if funcname is None else funcname
        raise RuntimeError(f"Error documenting {funcname}:\n{str(exp)}")

    return f


def _indentcount_lines(lines: List[str]) -> int:
    """Minimum indent for all lines in line list.

    >>> lines = [' one', '  two', '   three']
    >>> indentcount_lines(lines)
    1
    >>> lines = []
    >>> indentcount_lines(lines)
    0
    >>> lines = [' one']
    >>> indentcount_lines(lines)
    1
    >>> indentcount_lines(['    '])
    0
    """
    indent = sys.maxsize
    for line in lines:
        line_stripped = line.lstrip()
        if line_stripped:
            indent = min(indent, len(line) - len(line_stripped))
    if indent == sys.maxsize:
        return 0
    return indent


def copy_doc(source: Callable) -> Callable:
    """Copy the docstring from another function (decorator).

    The docstring of the source function is prepepended to the docstring of the
    function wrapped by this decorator.

    This is useful when inheriting from a class and overloading a method. This
    decorator can be used to copy the docstring of the original method.

    Parameters
    ----------
    source : callable
        The function to copy the docstring from.

    Returns
    -------
    wrapper : callable
        The decorated function.

    Examples
    --------
    >>> class A:
    ...     def m1():
    ...         '''Docstring for m1'''
    ...         pass
    >>> class B(A):
    ...     @copy_doc(A.m1)
    ...     def m1():
    ...         ''' this gets appended'''
    ...         pass
    >>> print(B.m1.__doc__)
    Docstring for m1 this gets appended
    """

    def wrapper(func):
        if source.__doc__ is None or len(source.__doc__) == 0:
            raise RuntimeError(
                f"The docstring from {source.__name__} could not be copied "
                "because it was empty."
            )
        doc = source.__doc__
        if func.__doc__ is not None:
            doc += func.__doc__
        func.__doc__ = doc
        return func

    return wrapper



import time

import numpy as np
import pytest

from stimuli.audio import BaseSound


def test_check_volume():
    """Check the volume static checker."""
    volume = BaseSound._check_volume(50)
    assert np.allclose(volume, np.array([50, 50]))
    volume = BaseSound._check_volume((25, 50))
    assert np.allclose(volume, np.array([25, 50]))
    volume = BaseSound._check_volume((0, 100))
    assert np.allclose(volume, np.array([0, 100]))
    with pytest.raises(
        TypeError, match="must be an instance of numeric or tuple"
    ):
        BaseSound._check_volume([25, 50])
    with pytest.raises(AssertionError):
        volume = BaseSound._check_volume((25, 50, 75))
    with pytest.raises(TypeError, match="must be an instance of numeric"):
        volume = BaseSound._check_volume(("25", 50))
    with pytest.raises(AssertionError):
        volume = BaseSound._check_volume((25, 101))
    with pytest.raises(AssertionError):
        volume = BaseSound._check_volume((-25, 100))


def test_check_sample_rate():
    """Check the sample rate static checker."""
    fs = BaseSound._check_sample_rate(44100)
    assert fs == 44100
    with pytest.raises(TypeError, match="must be an instance of int"):
        BaseSound._check_sample_rate(44100.0)
    with pytest.raises(AssertionError):
        BaseSound._check_sample_rate(-101)


def test_check_duration():
    """Check the duration static checker."""
    duration = BaseSound._check_duration(101)
    assert duration == 101
    duration = BaseSound._check_duration(101.01)
    assert duration == 101.01
    with pytest.raises(TypeError, match="must be an instance of numeric"):
        BaseSound._check_duration([101])
    with pytest.raises(AssertionError):
        BaseSound._check_sample_rate(-101)


def _test_base(Sound):
    """Test base functionalities with a Sound class."""
    sound = Sound(10, duration=0.2)

    # test play/stop
    sound.play()
    time.sleep(0.2)

    duration = 1
    sound = Sound(10, duration=duration)
    sound.play(blocking=True)

    start = time.time()
    sound.play(blocking=False)
    sound.stop()
    assert time.time() - start <= duration

    # test volume setter
    assert np.isclose(np.max(np.abs(sound.signal)), sound.volume / 100)
    sound.volume = 20
    assert sound.volume == 20
    assert np.isclose(np.max(np.abs(sound.signal)), sound.volume / 100)
    sound.volume = (20, 100)
    assert sound.volume == (20, 100)
    assert np.allclose(
        np.max(np.abs(sound.signal), axis=0), np.array(sound.volume) / 100
    )

    # test sample rate and duration setter
    assert sound.sample_rate == 44100
    assert sound.duration == duration
    assert sound.times.size == sound.sample_rate * sound.duration
    sound.sample_rate = 48000
    assert sound.sample_rate == 48000
    assert sound.duration == duration
    assert sound.times.size == sound.sample_rate * sound.duration
    sound.duration = 0.5
    assert sound.sample_rate == 48000
    assert sound.duration == 0.5
    assert sound.times.size == sound.sample_rate * sound.duration


def _test_no_volume(Sound):
    """Test signal if volume is set to 0."""
    sound = Sound(0)
    assert np.allclose(sound.signal, np.zeros(sound.signal.shape))



"""Test _checks.py"""

import logging

import pytest

from .._checks import _check_type, _check_value, _check_verbose, _ensure_int


def test_ensure_int():
    """Test _ensure_int checker."""
    # valids
    assert _ensure_int(101) == 101

    # invalids
    with pytest.raises(TypeError, match="Item must be an int"):
        _ensure_int(101.0)
    with pytest.raises(TypeError, match="Item must be an int"):
        _ensure_int(True)
    with pytest.raises(TypeError, match="Item must be an int"):
        _ensure_int([101])


def test_check_type():
    """Test _check_type checker."""
    # valids
    assert _check_type(101, ("int",)) == 101
    assert _check_type("101.fif", ("path-like",)) == "101.fif"

    def foo():
        pass

    _check_type(foo, ("callable",))

    assert _check_type(101, ("numeric",)) == 101
    assert _check_type(101.0, ("numeric",)) == 101.0

    # invalids
    with pytest.raises(TypeError, match="Item must be an instance of"):
        _check_type(101, (float,))
    with pytest.raises(TypeError, match="'number' must be an instance of"):
        _check_type(101, (float,), "number")


def test_check_value():
    """Test _check_value checker."""
    # valids
    assert _check_value(5, [1, 2, 3, 4, 5]) == 5
    assert _check_value((1, 2), [(1, 2), (2, 3, 4, 5)]) == (1, 2)

    # invalids
    with pytest.raises(ValueError, match="Invalid value for the parameter."):
        _check_value(5, [1, 2, 3, 4])
    with pytest.raises(
        ValueError, match="Invalid value for the 'number' parameter."
    ):
        _check_value(5, [1, 2, 3, 4], "number")


def test_check_verbose():
    """Test _check_verbose checker."""
    # valids
    assert _check_verbose(12) == 12
    assert _check_verbose("INFO") == logging.INFO
    assert _check_verbose("DEBUG") == logging.DEBUG
    assert _check_verbose(True) == logging.INFO
    assert _check_verbose(False) == logging.WARNING
    assert _check_verbose(None) == logging.WARNING

    # invalids
    with pytest.raises(TypeError, match="must be an instance of"):
        _check_verbose(("INFO",))
    with pytest.raises(ValueError, match="Invalid value"):
        _check_verbose("101")
    with pytest.raises(ValueError, match="negative integer, -101 is invalid."):
        _check_verbose(-101)
